(*
  Anonymous method completion plugin

  Copyright (c) 2015 Lyna

  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
    claim that you wrote the original software. If you use this software
    in a product, an acknowledgment in the product documentation would be
    appreciated but is not required.

    2. Altered source versions must be plainly marked as such, and must not be
    misrepresented as being the original software.

    3. This notice may not be removed or altered from any source
    distribution.
*)
unit UAnonComplete;

interface

uses
{$IFDEF DEBUG}
  Winapi.Windows,
{$ENDIF}
  System.SysUtils, System.Classes, System.RegularExpressions, Vcl.Menus, ToolsAPI;

type
  TKeyboardBinding = class(TNotifierObject, IOTAKeyboardBinding)
  private
    function GetBindingType: TBindingType;
    function GetDisplayName: string;
    function GetName: string;
    procedure BindKeyboard(const BindingServices: IOTAKeyBindingServices);

    procedure AnonMethodComplete(const Context: IOTAKeyContext; KeyCode: TShortcut;
      var BindingResult: TKeyBindingResult);
  end;

  TAnonMethod = record
    TypeName: string;
    Prototype: string;
  end;

const
  DefinedTypes: array[0..15] of TAnonMethod = (
    (TypeName: 'TBindCompDesignExpression.TExecuteDesignExpression'; Prototype: 'procedure(const AName: string; const AExpression: string; ACallback: TProc<IValue>)'),
    (TypeName: 'TBindCompDesignExpression.TExecuteDesignExpression2'; Prototype: 'procedure(const AName: string; const AControlExpression, ASourceExpression: string; ACallback: TProc<IValue>)'),
    (TypeName: 'TBindCompDesignExpression.TSaveDesignExpression'; Prototype: 'procedure(const AName: string; const AExpression: string)'),
    (TypeName: 'TCustomScope.TWrapperFoundCallback'; Prototype: 'procedure(Scope: TCustomScope; const Wrapper: ICustomWrapper)'),
    (TypeName: 'TLocationWrapper.TValueGetter'; Prototype: 'function: TValue'),
    (TypeName: 'TLocationWrapper.TValueSetter'; Prototype: 'procedure(AValue: TValue)'),
    (TypeName: 'TBindingCallbackFactory.TBindingCallback'; Prototype: 'function(const Obj: TObject; const Val: TValue): Boolean'),
    (TypeName: 'TGraph.TFindGraphNodeCallback'; Prototype: 'function(Node: TGraphNode): Boolean'),
    (TypeName: 'Log.ToStringFunc'; Prototype: 'function(O: TObject): string'),
    (TypeName: 'TObservers.TCanObserveEvent'; Prototype: 'function(const ID: Integer): Boolean'),
    (TypeName: 'TObservers.TObserverAddedEvent'; Prototype: 'procedure(const ID: Integer; const Observer: IObserver)'),
    (TypeName: 'TLoginCredentialService.TLoginFunc'; Prototype: 'function(const Username, Password, Domain: string): Boolean'),
    (TypeName: 'TList<T>.TEmptyFunc'; Prototype: 'function(const L, R: T): Boolean'),
    (TypeName: 'TList<T>.TListCompareFunc'; Prototype: 'function(const L, R: T): Integer'),
    (TypeName: 'TDirectory.TFilterPredicate'; Prototype: 'function(const Path: string; const SearchRec: TSearchRec): Boolean'),
    (TypeName: 'TDirectory.TDirectoryWalkProc'; Prototype: 'function(const Path: string; const FileInfo: TSearchRec): Boolean')
  );

implementation

{ TKeyboardBinding }

procedure TKeyboardBinding.AnonMethodComplete(const Context: IOTAKeyContext;
  KeyCode: TShortcut; var BindingResult: TKeyBindingResult);
var
  codeInsight: IOTACodeInsightServices;
  codeInsightMan: IOTACodeInsightManager;
  paramList: IOTACodeInsightParameterList;
  i, j, paramIndex: Integer;
  query: IOTACodeInsightParamQuery;
  arr: TArray<string>;
  m: TMatch;
  value: string;
  allow: Boolean;
  needClearing: Boolean;
  anonMethods: array of TAnonMethod;
  str: string;
  symbols: IOTACodeInsightSymbolList;
  dispParam: Boolean;
  row: Integer;

  function ExtractTypeName(const S: string): string;
  var
    p, p1: PChar;
    tmp: string;
    nest: Integer;
  begin
    Result := '';
    p := PChar(S);
    p1 := p;
    nest := 0;
    while p1^ <> #0 do
    begin
      case p1^ of
        '<':
        begin
          if nest = 0 then
          begin
            SetString(tmp, p, p1 - p);
            Result := Result + tmp + '<>';
          end;
          Inc(nest);
        end;
        '>':
        begin
          Dec(nest);
          if nest = 0 then
          begin
            p := p1 + 1;
          end;
        end;
      end;
      Inc(p1);
    end;
    if p <> p1 then
    begin
      SetString(tmp, p, p1 - p);
      Result := Result + tmp;
    end;
  end;

  function ExtractTypeParameters(const S: string): TArray<string>;
  var
    iStart, iEnd: Integer;
    types: string;
    p, p1: PChar;
    nest: Integer;
  begin
    Result := nil;

    iStart := S.IndexOf('<');
    iEnd := S.LastIndexOf('>');
    if (iStart = -1) or (iEnd = -1) then Exit;

    types := S.Substring(iStart + 1, iEnd - iStart - 1);
    p := PChar(types);
    p1 := p;
    nest := 0;
    while p1^ <> #0 do
    begin
      case p1^ of
        '<': Inc(nest);
        '>': Dec(nest);
        ',':
        begin
          if nest = 0 then
          begin
            SetLength(Result, Length(Result) + 1);
            SetString(Result[High(Result)], p, p1 - p);
            p := p1 + 1;
          end;
        end;
      end;
      Inc(p1);
    end;
    if p <> p1 then
    begin
      SetLength(Result, Length(Result) + 1);
      SetString(Result[High(Result)], p, p1 - p);
    end;
  end;

  function GetValueByName(const Name: string; out Value: string): Boolean;
  var
    i, j: Integer;
    typeParamNames, typeParamValues: TArray<string>;
    stripedValue: string;

    function RemoveUnitNames(const Value: string): string;
    begin
      Result := TRegEx.Replace(Value, '[_a-zA-Z][_a-zA-Z0-9]+\.', '');
    end;

  begin
    typeParamValues := ExtractTypeParameters(Name);

    if Length(typeParamValues) = 0 then
    begin
      for i := Low(anonMethods) to High(anonMethods) do
      begin
        if not Name.EndsWith(anonMethods[i].TypeName, True) then Continue;

        Value := anonMethods[i].Prototype;
        Exit(True);
      end;
    end
    else begin
      for i := Low(anonMethods) to High(anonMethods) do
      begin
        typeParamNames := ExtractTypeParameters(anonMethods[i].TypeName);
        if Length(typeParamNames) <> Length(typeParamValues) then Continue;
        if not ExtractTypeName(Name).EndsWith(ExtractTypeName(anonMethods[i].TypeName), True) then Continue;

        Value := anonMethods[i].Prototype;
        for j := Low(typeParamNames) to High(typeParamNames) do
        begin
          stripedValue := RemoveUnitNames(typeParamValues[j]);
          Value := TRegEx.Replace(Value, ': ' + typeParamNames[j] + '([;\)])', ': ' + stripedValue + '\1');
          Value := TRegEx.Replace(Value, '([,<])' + typeParamNames[j] + '([,>])', '\1' + stripedValue + '\2');
          Value := TRegEx.Replace(Value, ': ' + typeParamNames[j] + '$', ': ' + stripedValue);
          Value := RemoveUnitNames(Value);
        end;
        Exit(True);
      end;
    end;

    Result := False;
  end;

begin
  BindingResult := krUnhandled;

  codeInsight := BorlandIDEServices as IOTACodeInsightServices;
  codeInsight.GetCurrentCodeInsightManager(codeInsightMan);
  needClearing := False;

  if codeInsightMan = nil then
  begin
    for i := 0 to codeInsight.CodeInsightManagerCount-1 do
    begin
      if not codeInsight.CodeInsightManager[i].GetIDString.Contains('Pascal') then Continue;

      codeInsightMan := codeInsight.CodeInsightManager[i];
      codeInsight.SetQueryContext(Context.EditBuffer.TopView, codeInsightMan);
      needClearing := True;
      Break;
    end;

    if codeInsightMan = nil then
      Exit;
  end;
  try
    if not codeInsightMan.Enabled then Exit;

    Context.EditBuffer.EditPosition.Save;
    try
      Context.EditBuffer.EditPosition.MoveBOL;

      allow := True;
      codeInsightMan.AllowCodeInsight(allow, #0);
      if not allow then Exit;
      try
        str := '';
        if not codeInsightMan.InvokeCodeCompletion(itManual, str) then Exit;
        try
          row := Context.EditBuffer.EditPosition.Row;
          for i := row downto 1 do
          begin
            codeInsightMan.GetSymbolList(symbols);
            if symbols = nil then Continue;
            anonMethods := nil;
            for j := 0 to symbols.Count-1 do
            begin
              if symbols.SymbolTypeText[j].Contains('reference') then
              begin
                SetLength(anonMethods, Length(anonMethods) + 1);
                anonMethods[High(anonMethods)].TypeName := symbols.SymbolText[j];
                anonMethods[High(anonMethods)].Prototype := symbols.SymbolTypeText[j];
                anonMethods[High(anonMethods)].Prototype := anonMethods[High(anonMethods)].Prototype.Replace(': reference to ', '');
                anonMethods[High(anonMethods)].Prototype := anonMethods[High(anonMethods)].Prototype.Substring(0, anonMethods[High(anonMethods)].Prototype.Length-1);
              end;
            end;
            if Length(anonMethods) > 0 then
            begin
              for j := Low(DefinedTypes) to High(DefinedTypes) do
              begin
                SetLength(anonMethods, Length(anonMethods) + 1);
                anonMethods[High(anonMethods)].TypeName := DefinedTypes[j].TypeName;
                anonMethods[High(anonMethods)].Prototype := DefinedTypes[j].Prototype;
              end;
              Break;
            end;

            Context.EditBuffer.EditPosition.MoveRelative(-1, 0);
            Context.EditBuffer.EditPosition.MoveBOL;
            str := '';
            if not codeInsightMan.InvokeCodeCompletion(itManual, str) then Exit;
          end;
        finally
          dispParam := False;
          codeInsightMan.Done(False, dispParam);
        end;
      except
        Exit;
      end;
    finally
      Context.EditBuffer.EditPosition.Restore;
    end;

    allow := True;
    codeInsightMan.AllowCodeInsight(allow, #1);
    if not allow then Exit;
    try
      if not codeInsightMan.InvokeParameterCodeInsight(itManual, paramIndex) then Exit;
    except
      Exit;
    end;
    try
      codeInsightMan.GetParameterList(paramList);
      if paramList = nil then Exit;
      for i := 0 to paramList.ProcedureCount-1 do
      begin
        paramList.GetParameterQuery(i, query);
        if query = nil then Continue;

        arr := paramList.ProcedureParamsText[i].Split([#13#10], TStringSplitOptions.ExcludeEmpty);
        if paramIndex >= Length(arr) then Continue;

        m := TRegEx.Match(arr[paramIndex], '.*?\s?.*?: (.*)');
        if m.Success and GetValueByName(m.Groups[1].Value, value) then
        begin
          Context.EditBuffer.EditPosition.InsertText(#13#10);
          Context.EditBuffer.EditPosition.InsertCharacter(#9);
          Context.EditBuffer.EditPosition.InsertText(value+#13#10'begin'#13#10'end');
          BindingResult := krHandled;
          Exit;
        end;
      end;
    finally
      dispParam := False;
      codeInsightMan.Done(False, dispParam);
    end;
  finally
    if needClearing then
      codeInsight.SetQueryContext(nil, nil);
  end;
end;

procedure TKeyboardBinding.BindKeyboard(
  const BindingServices: IOTAKeyBindingServices);
begin
  BindingServices.AddKeyBinding([TextToShortCut('Ctrl+Alt+Space')], AnonMethodComplete, nil);
end;

function TKeyboardBinding.GetBindingType: TBindingType;
begin
  Result := btPartial;
end;

function TKeyboardBinding.GetDisplayName: string;
begin
  Result := 'Anonymous method completion';
end;

function TKeyboardBinding.GetName: string;
begin
  Result := 'AnonymousMethodCompletion';
end;

var
  KeyBindingIndex: Integer;
initialization
  KeyBindingIndex := (BorlandIDEServices as IOTAKeyboardServices).AddKeyboardBinding(TKeyboardBinding.Create);
finalization
  (BorlandIDEServices as IOTAKeyboardServices).RemoveKeyboardBinding(KeyBindingIndex);
end.
